"""Consolidate shared classes across entity types in acai/<entitytype>/shared/shared_classes.py.


The original code in acai/<entitytype>/shared/shared_classes.py has a
lot of duplication. This attempts to consolidate it into one place for
better maintainability.

This was constructed by copying the code from the various
entity-specific "shared/shared_classes.py" files into this file. I
checked people, places, deities, groups, and events, in that order. So
people/shared/shared_classes.py did most of the work. Note I did not
include the VizBible* classes or EntityDataEncode from that file.

Then i extracted common attributes into a base class AcaiEntity.

CAUTION: this means the original versions of these files could change,
until this gets synced with Rick's approach.

"""

from dataclasses import dataclass, field
import json as json
import re
from typing import Any, Optional

from biblelib.word import bcvwpid


# more preliminaries
@dataclass
class Localization:
    gloss: str
    short_definition: str = ""


@dataclass
class Identifier:
    identifier: str
    source: str


@dataclass
class LexicalSense:
    index: str
    source: str
    lemmas: list = field(default_factory=list)
    localizations: dict[str, Localization] = field(default_factory=dict)
    # glosses: list= field(default_factory=list)  # a list of LexicalGloss objects
    # short_definitions: list= field(default_factory=list) # a list of ShortDefinition objects


@dataclass
class EntityData:
    entity_class: str  # 'person' or 'place' or perhaps 'deity' (?)
    language: str  # but maybe a list because heb/aram?
    ubsdbh: str = ""
    ubsdgnt: str = ""
    lexical_sense: Optional[LexicalSense] = None
    lemmas: list = field(default_factory=list)
    segment_identifiers: dict[str, list] = field(default_factory=dict)  # a dict[edition, list<str>]
    pronominal_referents: dict[str, list] = field(
        default_factory=dict
    )  # a dict[edition, list<str>]
    subject_referents: dict[str, list] = field(default_factory=dict)  # a dict[edition, list<str>]


@dataclass
class AlternateSources:
    aquifer: list = field(default_factory=list)
    obi: list = field(default_factory=list)
    ubsdgnt: list = field(default_factory=list)
    ubsdbh: list = field(default_factory=list)
    ubsfauna: list = field(default_factory=list)
    ubsflora: list = field(default_factory=list)
    ubsrealia: list = field(default_factory=list)
    # biblemapper: list = field(default_factory=list)
    digital_atlas_roman_empire: list = field(default_factory=list)
    # using OBI factbook data does not work (logos changed schema?)
    # faithlife_factbook: list = field(default_factory=list)
    geonames: list = field(default_factory=list)
    pleiades: list = field(default_factory=list)
    tipnr: list = field(default_factory=list)
    vizbible: list = field(default_factory=list)
    wikidata: list = field(default_factory=list)
    wikipedia: list = field(default_factory=list)

    def __post_init__(self) -> None:
        """Ensure values are valid."""
        if self.aquifer:
            assert isinstance(self.aquifer, list), "aquifer must be a list"
            for item in self.aquifer:
                assert (
                    len(item.split("|")) == 3
                ), "aquifer items must be in the format 'id|name|source'"
        if self.ubsdgnt:
            assert isinstance(self.ubsdgnt, list), "ubsdgnt must be a list"
            # raw data combines multipl codes in a single string, so we split them
            self.ubsdgnt = [substr.strip() for item in self.ubsdgnt for substr in item.split()]
            for item in self.ubsdgnt:
                assert re.match(
                    r"^\d+\.\d+([a-z]|{N:001})?$", item
                ), "UBSDGNT code must match 'nn.nn' format"
        # others also need testing


@dataclass
class AcaiEntity:
    """Base class for all ACAI entities."""

    # type + identifier
    id: str
    # pointer to primary id. If this *is* the primary, it matches the id field
    primary_id: str
    alternate_sources: AlternateSources = field(default_factory=AlternateSources)
    # represents the entity class, e.g. "person", "place", "deity", "group", "event"
    # overwritten in subclasses
    type: str = ""
    # list of UBS DGNT identifiers
    ubsdgnt: list[str] = field(default_factory=list)
    # list of UBS DBH identifiers
    ubsdbh: list[str] = field(default_factory=list)
    # key is lang, dict inside has up to three entries: preferred label, alt labels, and description
    localizations: dict[str, dict[str, Any]] = field(default_factory=dict)
    # list of ids for entities that are possible the same as this one
    possibly_same_as: list[str] = field(default_factory=list)
    # list of ids pointing to respective entries
    referred_to_as: list[str] = field(default_factory=list)
    # True if only mentioned in the NRSV apocrypha
    only_mentioned_in_apocrypha: bool = False
    # True if not mentioned in the Bible
    non_biblical: bool = False
    # short list of the most important BCVID string references in ORG versification
    key_references: list = field(default_factory=list)
    # exhaustive list of BCVID string references in ORG versification
    references: list = field(default_factory=list)
    # dict[edition, list<BCVWPID>]
    explicit_instances: dict[str, list] = field(default_factory=dict)
    # dict[edition, list<BCVWPID>] for pronominal references
    pronominal_referents: dict[str, list] = field(default_factory=dict)
    # dict[edition, list<BCVWPID>] for subject references
    subject_referents: dict[str, list] = field(default_factory=dict)
    # dict[edition, list<BCVWPID>] if there are any entities of this type that speak
    speeches: dict[str, list] = field(default_factory=dict)
    # lang: list<str> (list of lemmas in lang)
    lemmas: dict[str, list] = field(default_factory=dict)
    _valid_types: tuple[str, ...] = (
        "realia",
        "flora",
        "fauna",
        "person",
        "place",
        "deity",
        "group",
        "event",
    )

    def __post_init__(self) -> None:
        """Ensure values are valid."""
        if not isinstance(self.id, str):
            raise TypeError(f"id must be a string, got {type(self.id)}")
        if not isinstance(self.primary_id, str):
            raise TypeError(f"primary_id must be a string, got {type(self.primary_id)}")
        assert (
            self.type in self._valid_types
        ), f"type must be one of {self._valid_types}, got {self.type}"
        if self.ubsdgnt:
            assert isinstance(self.ubsdgnt, list), "ubsdgnt must be a list"
            # raw data combines multipl codes in a single string, so we split them
            self.ubsdgnt = [substr.strip() for item in self.ubsdgnt for substr in item.split()]
            for item in self.ubsdgnt:
                assert re.match(
                    r"^\d+\.\d+([a-z]|{N:001})?$", item
                ), "UBSDGNT code must match 'nn.nn' format"
        if self.localizations and "eng" in self.localizations:
            assert isinstance(
                self.localizations["eng"], dict
            ), "localizations['eng'] must be a dict"
            assert (
                "preferred_label" in self.localizations["eng"]
            ), "localizations['eng'] must contain 'preferred_label'"
        if self.referred_to_as:
            assert isinstance(self.referred_to_as, list), "referred_to_as must be a list"
            for item in self.referred_to_as:
                assert ":" in item, "referred_to_as items must be in the format 'type:id'"
        assert isinstance(
            self.only_mentioned_in_apocrypha, bool
        ), "only_mentioned_in_apocrypha must be a boolean"
        assert isinstance(self.non_biblical, bool), "non_biblical must be a boolean"

    def __repr__(self) -> str:
        """String representation of the entity."""
        return f"Acai{self.type.capitalize()}(id={self.id}, primary_id={self.primary_id})"

    def preflabel(self, language: str = "eng") -> str:
        """Return the preferred label in the specified language."""
        if language != "eng":
            raise NotImplementedError("Not implemented for languages other than English.")
        localizations = self.localizations.get(language, {})
        if localizations:
            return localizations.get("preferred_label", "")

    def identifying_str(self, language: str = "eng") -> str:
        """Return an identifying string."""
        if language != "eng":
            raise NotImplementedError("Not implemented for languages other than English.")
        entitystrmap = {
            "realia": "thing",
            "flora": "plant, flower, or tree",
            "fauna": "animal",
            "person": "person",
            "place": "place or location",
            "deity": "deity or god",
            "group": "group of people or organization",
            "event": "event",
        }
        entitystr = entitystrmap.get(self.type, "entity")
        key_references_usfm: list[str] = ", ".join(
            [bcvwpid.BCVID(ref).to_usfm() for ref in self.key_references]
        )
        bibstring = "biblical" if not self.non_biblical else "non-biblical"
        return (
            f"{self.preflabel()} is a {bibstring} {entitystr} mentioned in {key_references_usfm}."
        )


@dataclass(repr=False)
class AcaiRealia(AcaiEntity):
    type: str = "realia"
    realia_type: str = "other"


@dataclass(repr=False)
class AcaiFlora(AcaiEntity):
    type: str = "flora"
    flora_type: str = "other"


@dataclass(repr=False)
class AcaiFauna(AcaiEntity):
    type: str = "fauna"
    # save the fauna_type until UBS fauna dict becomes open
    fauna_type: str = "other"


@dataclass(repr=False)
class AcaiGroup(AcaiEntity):
    type: str = "group"
    group_type: str = (
        "other"  # default is 'other'; other values: 'residence', 'relation', or 'religious'
    )
    group_origin: str = ""  # id of person or place group derives from, when known


@dataclass(repr=False)
class AcaiDeity(AcaiEntity):
    type: str = "deity"
    deity_type: str = (
        "deity"  # could be 'angel' or 'demon' or 'deity' or 'other'. 'deity' is default
    )
    # gender: str = "" # 'male', 'female', 'deity'; # vizbible
    # birth_place: str = "" # vizbible
    # death_place: str = "" # vizbible
    # father: str = "" # vizbible 'father'
    # mother: str = "" # vizbible 'mother'
    # partners: list = field(default_factory=list) # vizbible 'partners'
    # offspring: list = field(default_factory=list) # vizbible 'children'
    # siblings: list = field(default_factory=list) # vizbible
    related_places: dict[str, list] = field(default_factory=dict)  # reason, list of place:id
    # mentioned_in_bible: bool = True # "bible" == entire NRSV (ot, nt, apocrypha)


@dataclass(repr=False)
class AcaiPerson(AcaiEntity):
    type: str = "person"
    gender: str = ""  # 'male', 'female', 'deity'; # vizbible
    birth_place: str = ""  # vizbible
    death_place: str = ""  # vizbible
    father: str = ""  # vizbible 'father'
    mother: str = ""  # vizbible 'mother'
    partners: list = field(default_factory=list)  # vizbible 'partners'
    offspring: list = field(default_factory=list)  # vizbible 'children'
    siblings: list = field(default_factory=list)  # vizbible
    roles: list = field(default_factory=list)  # curation
    tribe: str = ""  # tipnr 'tribe' mapped to ACAI group identifier
    related_places: dict[str, list] = field(default_factory=dict)  # reason, list of place:id
    # associated_places: list = field(default_factory=list) # list of acai_id
    # mentioned_in_bible: bool = True # "bible" == entire NRSV (ot, nt, apocrypha)
    # is_artifact: bool = False
    # is_person: bool = False
    # added_entry: bool = False
    # will there be person types? roles?
    # place_types: dict[str, list] = field(default_factory=dict) # key is source, list of types in source


@dataclass(repr=False)
class AcaiPlace(AcaiEntity):
    type: str = "place"
    tribal_area: str = ""
    associated_places: list = field(default_factory=list)  # list of acai_id
    nearby_places: list = field(default_factory=list)  # list of acai_id
    subregion_of: list = field(default_factory=list)  # list of acai_id
    mentioned_in_bible: bool = True  # "bible" == entire NRSV (ot, nt, apocrypha)
    is_artifact: bool = False
    is_person: bool = False
    added_entry: bool = False
    place_types: dict[str, list] = field(
        default_factory=dict
    )  # key is source, list of types in source
    geocoordinates: dict[str, dict[str, str]] = field(default_factory=dict)  # unsure of type


@dataclass
class LocationInfo:
    obi_id: str
    obi_friendly_id: str
    obi_names: list = field(default_factory=list)
    # obi = openbibleinfo
    obi_types: list = field(default_factory=list)
    obi_references: list = field(default_factory=list)
    # add possibly_same_as ? Are these even needed if there are SDBH and SDBG alternate entries?
    ubsdgnt: list = field(default_factory=list)
    ubsdbh: list = field(default_factory=list)
    # lists or objects?
    digital_atlas_roman_empire: list = field(default_factory=list)
    # using OBI factbook data does not work (logos changed schema?)
    # faithlife_factbook: list = field(default_factory=list)
    pleiades: list = field(default_factory=list)
    tipnr: list = field(default_factory=list)
    wikidata: list = field(default_factory=list)
    wikipedia: list = field(default_factory=list)



"""Read a plaintext TSV file produced by src/article2tsv.sh.

This clones/diverges from TyndaleBibleDictionary/tynbd/Reader.py: i
should make that or a derivative into a library.

>>> from acai.util import TyndaleReader
>>> tynrd = TyndaleReader.Reader()
>>> len(tynrd)
5027
>>> list(tynrd.items())[23]
('AbelShittim', Article(identifier='AbelShittim', headword='ABEL-SHITTIM', paragraphs=['Alternate name for Shittim, a place on\xa0the plains of Moab, in bref^Num_33_49. See art^ShittimPlace.'], wordcount=14, isredirect=True, contenttype=''))

"""

from collections import UserDict, Counter
from dataclasses import dataclass, field
from pathlib import Path
import re
from statistics import median

TYNDATAPATH = Path(__file__).parent.parent.parent.parent / "TyndaleBibleDictionary/plaintext"


@dataclass
class Article:
    identifier: str
    headword: str
    paragraphs: list[str] = field(default_factory=list)
    wordcount: int = 0
    isredirect: bool = False
    contenttype: str = ""

    def __post_init__(self) -> None:
        """Compute values after initialization."""
        # count 'words': space-delimited tokens
        self.wordcount = len([w for p in self.paragraphs for w in p.split(" ")])
        # Phrases like '(Alternate|Shortened) (name|form) (for|of)',
        # 'RSV rendering of', 'Name of ____ in', etc. These are redirects but it's also
        # worth capturing the alternate.
        # self.alternatename: str
        # could do more with this heuristic, which overgenerates a
        # bit: 'Ab' has little content but isn't a complete
        # redirect.
        self.isredirect = ("See art^" in self.paragraphs[0]) and len(self.paragraphs) == 1

    @property
    def clean_headword(self) -> str:
        """Return a cleaned up headword string."""
        cleaned = self.headword
        if cleaned.endswith("*"):
            cleaned = cleaned[:-1]
        return cleaned

    def write(self, outpath: Path) -> None:
        """Write article to outpath."""
        with outpath.open("w") as f:
            f.write(self.headword + "\n")
            for p in self.paragraphs:
                f.write(p + "\n")

    def numbered_paragraphs(self) -> list[str]:
        """Return any paragraphs that start with a number.

        This is a pattern for articles that need disambiguation.
        """
        expr = re.compile(" ?[1-9]+\\. ")
        return [para for para in self.paragraphs if expr.match(para)]

    def is_numbered(self, ratio: float = 0.5) -> bool:
        """Return True if more paragraphs are numbered than the ratio."""
        if len(self.paragraphs) == 0:
            return False
        else:
            return (len(self.numbered_paragraphs()) / len(self.paragraphs)) > ratio


class Reader(UserDict):
    """Read the articles from a TSV file."""

    filestems = [
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
        "G",
        "H",
        "I",
        "J",
        "K",
        "L",
        "M",
        "N",
        "O",
        "P",
        "Q",
        "R",
        "S",
        "T",
        "U",
        "V",
        "W",
        # there's always one outlier :-/
        "XY",
        "Z",
    ]

    def __init__(self) -> None:
        """Initialize an instance."""
        super().__init__()
        for stem in self.filestems:
            filepath = TYNDATAPATH / f"{stem}.tsv"
            try:
                self.read_file(tsvpath=filepath)
            except Exception as e:
                print(f"Failed on {filepath}\n{e}")
        self.wordcount = sum(art.wordcount for art in self.values())
        self.median_wordcount = median

    def read_file(self, tsvpath: Path) -> None:
        """Read a TSV file into memory."""
        self.tsvpath = tsvpath
        with self.tsvpath.open() as f:
            lastid: str = ""
            headword: str = ""
            paragraphs: list[str] = []
            for line in f.readlines():
                try:
                    cleanline = line.strip()
                    # a handful don't have content: need to fix upstream
                    if "\t" not in cleanline:
                        continue
                    artid, paragraph = line.strip().split("\t")
                    if artid != lastid:
                        # new article: output previous data, reset
                        if lastid:
                            self.data[lastid] = Article(
                                identifier=lastid, headword=headword, paragraphs=paragraphs
                            )
                        # first line has the headword
                        headword = paragraph
                        paragraphs = []
                        lastid = artid
                    else:
                        # continue previous article
                        paragraphs.append(paragraph)
                except Exception as e:
                    print(f"Failed on {line}\n{e}")

    def wordcount_bins(self, binsize: int = 100) -> Counter:
        """Return a counter of article wordcounts by binsize."""
        return Counter(int(art.wordcount / binsize) for art in self.values())

    def numbered_paragraphs(self) -> list[Article]:
        """Return a list of numbered paragraphs."""
        return [art for art in self.values() if art.is_numbered()]



"""Common code for rendering user displays of ACAI entity data (people, places, deities, groups).

This assumes the dataclasses in shared/shared_classes.
"""

from dataclasses import dataclass, field
import regex as re
from typing import Any

from biblelib.word import bcvwpid


@dataclass
class Entity:
    """A base class for rendering ACAI entities.

    This abstracts the shared_classes dataclasses for rendering via
    Flask and Jinja2 templates. The display is organized for English.

    """

    entity: Any
    id: str
    # pointer to primary id. If this *is* the primary, it matches the id field
    primary_id: str
    type: str
    subtype: str = ""
    preferred_label: str = ""
    alternate_labels: list = field(default_factory=list)
    # a list of acai_id pointing to respective entries
    referred_to_as: list = field(default_factory=list)
    # "apocrypha" == NRSV apocrypha only
    only_mentioned_in_apocrypha: bool = False
    # "non-biblical" == not in the Bible at all
    non_biblical: bool = False
    # key is description source, value is the description string
    descriptions: dict[str, str] = field(default_factory=dict)
    hebrew_lemmas: list = field(default_factory=list)
    greek_lemmas: list = field(default_factory=list)
    ubsdgnt: list = field(default_factory=list)
    ubsdbh: list = field(default_factory=list)
    key_references: list = field(default_factory=list)
    references: list = field(default_factory=list)
    # type-specific additional fields, as label/value pairs
    # must be formatted here so the template doesn't have to know about them
    extensions: dict[str, str] = field(default_factory=dict)

    def __repr__(self) -> str:
        """Return a printed representation."""
        return f"<Entity: '{self.id}'>"

    @staticmethod
    def from_dataclass(entity: Any) -> "Entity":
        """Return an Entity instance from an ACAI dataclass."""

        # core functionality here ought to live in Biblelib
        def to_usfm(reflist: list[str]) -> list[str]:
            """Return references in USFM format."""
            return [bcvwpid.make_id(ref).to_usfm() for ref in reflist]

        def clean_description(description: str) -> str:
            """Return a cleaned-up description string.

            Removes HTML tags and special characters.

            """
            # regex to remove <..>
            clean_description: str = re.sub(r"<[^>]+>", "", description)
            clean_description = re.sub(r"\{[A-Z]+:([^}]+)}", r"\1", clean_description)
            clean_description = re.sub(r" *[◄►] *", " ", clean_description)
            clean_description = re.sub(r"\.\s*$", "", clean_description)
            return clean_description

        # doesn't look liek this gets the second level of
        # interpretation by Flask: probably another way
        # Also needs to handle _lists_ of identifiers.
        # def url_if_identifier(string: str) -> str:
        #     """Return an a element if string is an ACAI identifier.

        #     Otherwise just the string value.

        #     """
        #     entitytypes = {"deity", "group", "person", "place"}
        #     if ":" in string:
        #         typestr, _ = string.split(":")
        #         if typestr in entitytypes:
        #             urlforstr: str = f"url_for('identifierlookup', identifier='{string}')"
        #             return '<a href="{{  ' + urlforstr + '}}">'
        #         else:
        #             return string
        #     else:
        #         return string

        eng_localizations = entity.localizations.get("eng", {})
        eng_preferred_label: str = ""
        eng_alternate_labels: list[str] = []
        if eng_localizations:
            eng_descriptions = {
                desc["source"]: clean_description(desc["description"])
                for desc in eng_localizations.get("descriptions", {})
            }
            eng_preferred_label = eng_localizations.get("preferred_label", "")
            eng_alternate_labels = eng_localizations.get("alternate_labels", [])
        # add any type-specific extensions. Key is formatted for
        # display: values is used to retrieve the entity attribute.
        extensionsmap: dict[str, dict[str, str]] = {
            "deity": {
                "Related places": "related_places",
            },
            "group": {
                "Group origin": "group_origin",
            },
            "person": {
                "Gender": "gender",
                "Birthplace": "birth_place",
                "Death place": "death_place",
                "Father": "father",
                "Mother": "mother",
                "Partners": "partners",
                "Offspring": "offspring",
                "Siblings": "siblings",
                "Roles": "roles",
                "Related places": "related_places",
                "Possibly same as": "possibly_same_as",
            },
            "place": {
                "Possibly same as": "possibly_same_as",
                "Associated places": "associated_places",
                "Nearby places": "nearby_places",
                "Subregion of": "subregion_of",
            },
        }
        extensions: dict[str, str] = {
            label: attrval
            for label, attr in extensionsmap.get(entity.type, {}).items()
            if (attrval := getattr(entity, attr))
            if attrval
        }
        # need more to capture speeches for People
        subtype: str = ""
        if entity.type == "deity":
            subtype = entity.deity_type
        elif entity.type == "group":
            subtype = entity.group_type
        elif entity.type == "place":
            subtype = entity.place_types.get("acai", "")
        # elif entity.type == "person":
        #     subtype = entity.place_types.get("acai", "")
        return Entity(
            entity=entity,
            id=entity.id,
            primary_id=entity.primary_id,
            type=entity.type,
            subtype=subtype,
            # subtype=entity.subtype,
            preferred_label=eng_preferred_label,
            alternate_labels=eng_alternate_labels,
            referred_to_as=entity.referred_to_as,
            only_mentioned_in_apocrypha=entity.only_mentioned_in_apocrypha,
            non_biblical=entity.non_biblical,
            descriptions=eng_descriptions,
            hebrew_lemmas=entity.lemmas.get("he", {}),
            greek_lemmas=entity.lemmas.get("el", {}),
            ubsdbh=entity.ubsdbh,
            ubsdgnt=entity.ubsdgnt,
            key_references=to_usfm(entity.key_references),
            references=to_usfm(entity.references),
            extensions=extensions,
        )



"""Reader for Macula Greek data for SBLGNT.

This enables a variety of lookups.
"""

from collections import defaultdict, UserDict
import pandas as pd


def normalize_strongs(i: int, prefix: str = "G") -> str:
    """Return a normalized Strongs id for an int."""
    assert prefix in ["G", "H"], f"Prefix should be 'G' or 'H': {prefix}"
    istr = str(i)
    assert len(istr) <= 4, f"Too many digits: {istr}"
    return f"{prefix}{i:0>4}"


class _BaseReader(UserDict):
    """Base class for reading Macula TSV data from Github."""

    # placeholders
    macula_path = ""
    prefix = "X"
    # column label for Louw-Nida identifier
    senseattr = "ln"
    # column label for Strong's numbers
    strongattr = "strong"
    # pos not in this list are filtered out when reading
    keeppos = set(("adj", "adv, noun", "verb"))

    def __init__(self) -> None:
        """Initialize a Reader."""
        super().__init__()
        self.df = pd.read_csv(self.macula_path, delimiter="\t")
        self.strong_lemma = defaultdict(set)
        # maps a normalized strongs number to a set of LN sense ids
        # roughly 30% of strongs numbers have multiple senses
        # strongs numbers
        self.strong_ln = defaultdict(set)
        for index, row in self.df.iterrows():
            if getattr(row, "class") in self.keeppos:
                self.data[getattr(row, "xml:id")] = row
                strong = normalize_strongs(getattr(row, self.strongattr), prefix=self.prefix)
                self.strong_lemma[strong].add(row.lemma)
                self.strong_ln[strong].add(getattr(row, self.senseattr))


# Hebrew TSV is borked as of 2023-12-20: read Nodes data instead
# class HebrewReader(_BaseReader):
#     """Manage WLC TSV data read from Github."""

#     macula_path = "https://github.com/Clear-Bible/macula-hebrew/raw/main/TSV/macula-hebrew.tsv"
#     prefix = "H"
#     # column label for Strong's numbers
#     strongattr = "strongnumberx"


class GreekReader(_BaseReader):
    """Manage SBLGNT TSV data read from Github."""

    macula_path = (
        "https://github.com/Clear-Bible/macula-greek/raw/main/SBLGNT/tsv/macula-greek-SBLGNT.tsv"
    )
    prefix = "G"



"""Read the UBS dictionaries into lookup dicts.

The dataclass reflect the XML hierarchy (selectively):
Lexicon_Entry.baseforms ->
BaseForm.lexMeanings ->
LEXMeaning.lexSenses ->
LEXSense

# iterate over all the senses
>>> senses = [lexsense for lexent in ubsprs.values()
 for baseform in lexent.baseforms
 for lexmean in baseform.lexMeanings
 for lexsense in lexmean.lexSenses]
>>> len(senses)
9166
>>> senses[23].asdict()
{'languageCode': 'en', 'glosses': ['good', 'goodness', 'good act'], 'definitionShort': 'positive moral qualities of the most general nature', 'definitionLong': '', 'comments': []}

"""

from dataclasses import dataclass, field
from typing import Any, Optional

from biblelib.word import bcvwpid


@dataclass
class LEXSense:
    """Represents the same element from UBSGNT as a Python dataclass."""

    # no distinct identifier: 1-1 with LEXMeaning for Greek
    # the language of the definition
    languageCode: str
    glosses: list[str]
    definitionShort: str
    definitionLong: str
    comments: list[str]
    # link upward: beware recursion
    parent: Optional["LEXMeaning"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<LEXSense: {';'.join(self.glosses)}>"

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        fields = ["languageCode", "glosses", "definitionShort", "definitionLong", "comments"]
        return {f: getattr(self, f) for f in fields}


@dataclass
class LEXMeaning:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    isBiblicalTerm: str
    entryCode: str
    # only a single value for DGNT
    lexDomains: list[str]
    # only a single value for DGNT
    lexSubDomains: list[str]
    # appears to only be one LEXSense per LEXMeaning? But the
    # structure allows for more
    lexSenses: list[LEXSense]
    # using MARBLE-style references
    # this includes the deuterocanon, which seems to often have
    # multiple meanings for a single reference
    # some unusual ones like '04500900300002{N:002}'
    lexReferences: list[str]
    # link upward: beware recursion
    parent: Optional["BaseForm"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<LEXMeaning: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        fields = ["identifier", "isBiblicalTerm", "entryCode", "lexDomains", "lexSubDomains"]
        return {f: getattr(self, f) for f in fields}

    def serialize(self) -> dict[str, str]:
        """Return content as dict for serialization.

        Flattens single-value lists into strings, and omits references and parent.
        """
        refs: list[str] = []
        # the first three should be enough
        for ref in self.lexReferences[:3]:
            try:
                if len(ref) == 12:
                    refs.append(bcvwpid.BCVWPID(ref).to_usfm(with_word=True))
                else:
                    refs.append(bcvwpid.BCVID(ref).to_usfm())
            except Exception as e:
                print(f"Error processing reference {ref} from {self.identifier}: {e}")
                refs.append(ref)
        return {
            "identifier": self.identifier,
            "isBiblicalTerm": self.isBiblicalTerm,
            "lexDomains": ";".join(self.lexDomains),
            "lexSubDomains": ";".join(self.lexSubDomains),
            "entryCode": self.entryCode,
            "lemma": self.parent.parent.lemma,
            "glosses": ";".join(self.lexSenses[0].glosses),
            "definitionShort": self.lexSenses[0].definitionShort,
            "lexReferences": ";".join(refs),
        }

    def serialize_text(self) -> dict[str, str]:
        """Return textual content as dict for serialization.

        Flattens single-value lists into strings, and omits references and parent.
        """
        return {
            "entryCode": self.entryCode,
            "lemma": self.parent.parent.lemma,
            "glosses": ";".join(self.lexSenses[0].glosses),
            "definitionShort": self.lexSenses[0].definitionShort,
            "lexDomains": ";".join(self.lexDomains),
            "lexSubDomains": ";".join(self.lexSubDomains),
        }


@dataclass
class BaseForm:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    partsofspeech: list[str]
    lexMeanings: list[LEXMeaning]
    # FUTURE: incorporate RelatedLemmas.L
    # link upward: beware recursion
    parent: Optional["Lexicon_Entry"] = field(init=False, default=None)

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<BaseForm: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def asdict(self) -> dict[str, Any]:
        """Return content as dict, minus the parent to avoid infinite recursion."""
        selfdict = {f: getattr(self, f) for f in ["identifier", "partsofspeech"]}
        # if self.parent:
        #     # need to populate these for testing
        #     selfdict["lemma"] = self.parent.lemma
        #     # yuck but ...
        #     selfdict["strongCodes"] = self.parent.strongCodes
        return selfdict


@dataclass
class Lexicon_Entry:
    """Represents the same element from UBSGNT as a Python dataclass."""

    identifier: str
    lemma: str
    strongCodes: list[str]
    # FUTURE: capture Notes (see Lexicon_Entry Id="003076000000000")
    # multiple values if different parts of speech,
    # e.g. <Lexicon_Entry: 000266000000000>
    baseforms: list[BaseForm]

    def __repr__(self) -> str:
        """Return a string representation of the object."""
        return f"<Lexicon_Entry: {self.identifier}>"

    def __hash__(self) -> int:
        """Return a hash value."""
        return hash(self.identifier)

    def get_meanings(self) -> list[LEXMeaning]:
        """Return a list of of LEXMeanings for this entry."""
        return [lm for bf in self.baseforms for lm in bf.lexMeanings]



"""Compute importance scores for people.

>>> from acai.core import peopleimportance
# this uses a magic formula to compute the importance score
>>> peoplei = peopleimportance.PeopleImportance()
>>> peoplei.entitydf.loc["person:Mary"]
family                  0.572278
is_nt                   1.000000
n_book_references       0.085106
n_chapter_references    0.049180
n_key_references        0.115385
n_verse_references      0.007761
roles                   0.093750
importance              0.131374
Name: person:Mary, dtype: float64

# write out the scores to a TSV file for easy access. If you upgrade
# the magic formula, you should probably rewrite the TSV
>>> peoplei.write_tsv()


"""

# other ideas:
# - a person is more important if their family members are important
#   (e.g. "person:Mary")

from collections import defaultdict

import numpy as np
import pandas as pd

from acailib.shared_classes import AcaiPerson
from acailib.util import EntityReader

from .baseimportance import BaseImportance


class PeopleImportance(BaseImportance):
    """Compute importance scores for people."""

    reader = EntityReader.PeopleReader()
    type = "people"

    def __init__(self) -> None:
        """Initialize the importance score."""
        super().__init__()
        self.factors = {
            # probably don't want to even consider entities that aren't biblical at all
            # "biblical": 0.05,
            "family": 0.05,
            "is_nt": 0.05,
            # ditto for biblical
            # "nonapochryphal": 0.05,
            "n_book_references": 0.15,
            # would be nice to have pericope references but chapters are easier to compute
            "n_chapter_references": 0.15,
            "n_key_references": 0.10,
            "n_verse_references": 0.30,
            "roles": 0.20,
        }
        self.__post_init__()
        # person:Simeon -> person:Peter (but not vice versa)
        self.directed_aliases: dict[str, str] = {}
        # person:Peter -> {person:Simeon, person:Peter, etc.}: all the aliases
        self.aliases: dict[str, set[str]] = defaultdict(set)
        for ent in self.reader.values():
            if ent.primary_id != ent.id:
                if ent.id in self.directed_aliases:
                    print(f"Duplicate alias {ent.id} for {ent.primary_id}")
                self.directed_aliases[ent.id] = ent.primary_id
                # this will collect _all_ the aliases for a given entity
                self.aliases[ent.primary_id].add(ent.primary_id)
                self.aliases[ent.primary_id].add(ent.id)
        # only work with the primary record: in theory we're
        # collecting all the related data for aliases below
        self.primarykeys = [k for k in self.reader.keys() if k not in self.directed_aliases]
        self.entitydf = pd.DataFrame(index=self.primarykeys)
        # add all the factors to the dataframe, normalizing the range
        for factor in self.factors:
            self.entitydf[factor] = self.normalize_to_unit_range(
                np.asarray([getattr(self, factor)(self.reader[pid]) for pid in self.entitydf.index])
            )
        # compute the overall importance score
        self.entitydf["importance"] = self.entitydf.apply(self.compute_importance, axis=1)

    def _get_aliased_entries(self, entry: AcaiPerson) -> list[AcaiPerson]:
        """Get all the entries for a given entity."""
        primaryid: str = entry.primary_id
        if entry.id in self.directed_aliases:
            primaryid = self.directed_aliases[entry.id]
        if primaryid in self.aliases:
            return [self.reader[alias] for alias in self.aliases[primaryid]]
        else:
            return [entry]

    # this might overemphasize OT people with extensive family mentions
    def family(self, entry: AcaiPerson) -> float:
        """Compute the weighted number of known family members.

        The intuition is that the number of family members mentioned
        for a person indicates something about their importance.

        Parents and offspring are weighted more heavily than siblings.

        """
        weight: float = 0.0
        # this assumes that all aliased entries have the same family attributes
        for close_relative in ["father", "mother", "partners", "tribe"]:
            if getattr(entry, close_relative):
                weight += 1
        if entry.siblings:
            # the decremented weight of the number of siblings
            weight += self._decrement_length_weight(entry.siblings) / 2
        if entry.offspring:
            # the decremented weight of the number of offspring
            weight += self._decrement_length_weight(entry.offspring) / 2
        return weight

    def roles(self, entry: AcaiPerson) -> float:
        """Compute the aggregate weight for the number of roles for a person.

        The more roles a person has, the more important they are.
        """
        roleweights = {
            "Angel": 5,
            "Apostle": 5,
            "Army Commander": 2,
            "Carpenter": 3,
            "Church Leader": 4,
            "Concubine": 1,
            "Deity": 5,
            "Disciple": 4,
            "Emperor": 3,
            "Evangelist": 3,
            "Fisher": 4,
            "Fisherman": 4,
            "God": 5,
            "Governor": 3,
            "High Priest": 3,
            "Hunter": 1,
            "Judge": 3,
            "King": 4,
            "Lawgiver": 1,
            "Lawyer": 1,
            "Leader": 2,
            "Martyr, ": 3,
            "Matriarch": 1,
            "Member of the Royal Family": 2,
            "Messiah": 5,
            "Military Commander": 2,
            "Missionary": 3,
            "Official": 3,
            "Patriarch": 3,
            "Priest": 3,
            "Prisoner": 2,
            "Prophet": 4,
            "Queen": 4,
            "Relative of Jesus": 3,
            "Royal Family Member": 3,
            "Ruler": 3,
            "Ruler of the Synagogue": 2,
            "Scribe": 2,
            "Servant": 2,
            "Shepherd": 1,
            "Singer": 1,
            "Slave": 1,
            "Son of Adam": 3,
            "Tax Collector": 2,
            "Teacher": 3,
            "Tent Maker": 1,
            "Tribal Chief": 3,
            "Warrior": 2,
            "Wife": 1,
        }
        aliased_entries = self._get_aliased_entries(entry)
        allroles = {role for ent in aliased_entries for role in ent.roles}
        return sum(roleweights.get(role, 0) for role in allroles)



"""Read ACAI data into a dict.

>>> from acailib import EntityReader as er
>>> peoplerd = er.PeopleReader()
# return a list of entities whose preferred label matches a given string
>>> peoplerd.eng_preferred_label_dict["John"]
['person:John.3', 'person:John', 'person:John.2', 'person:John.5', 'person:John.4']
# likewise for other EntityReader subclasses

"""

from collections import UserDict, defaultdict
from collections.abc import Callable
import json
from pathlib import Path
from typing import Optional

from acailib import ACAIROOT
from acailib.shared_classes import AcaiDeity, AcaiGroup, AcaiPerson, AcaiPlace


class EntityReader(UserDict):
    """Base class for reading entity data."""

    acaidir: str = ""
    jsondir: Path = Path()
    # typing here needs help
    entryclass: Optional[Callable] = None
    repodata: str = "https://github.com/Clear-Bible/ACAI/tree/main/data"

    def __init__(self, upcase: bool = False) -> None:
        """Initialize an instance.

        With upcase = True (default is False), convert names to all
        upper-case. This is useful for matching TynBD headwords that
        are all upper-case.

        """
        super().__init__()
        for jsonfile in self.jsondir.glob("*.json"):
            with jsonfile.open() as f:
                jdict = json.load(f)
            try:
                if self.entryclass:
                    self.data[jdict["id"]] = self.entryclass(**jdict)
            except Exception as e:
                print(f"Failed on {jsonfile}\n{e}")
        self.eng_preferred_label_dict: defaultdict = defaultdict(list)
        self.eng_alternate_label_dict: defaultdict = defaultdict(list)
        for entityid in self.data:
            # split is a hack to dropo post-name qualifiers
            preflabel = self[entityid].localizations["eng"]["preferred_label"].split(" ")[0]
            altlabels = self[entityid].localizations["eng"].get("alternate_labels", [])
            if upcase:
                preflabel = preflabel.upper()
                altlabels = [label.upper() for label in altlabels]
            self.eng_preferred_label_dict[preflabel].append(entityid)
            for label in altlabels:
                self.eng_alternate_label_dict[label].append(entityid)

    def acai_uri(self, identifier: str) -> str:
        """Return a string for the Github repo location for identifier."""
        nsp, name = identifier.split(":")
        return f"{self.repodata}/{self.acaidir}/json/acai/{name}.json"

    def acai_md_link(self, identifier: str) -> str:
        """Return a Markdown link to the JSON file for identifier."""
        uri = self.acai_uri(identifier)
        return f"[{identifier}]({uri})"

    def references_by_type(self, identifier: str) -> dict[str, list[str]]:
        """Return all the references for an entity, grouped by type."""
        reftypes = [
            # note these overlap with other reference types
            "key_references",
            "references",
            "explicit_instances",
            "pronominal_referents",
            "subject_referents",
        ]
        return {reftype: getattr(self[identifier], reftype) for reftype in reftypes}

    def word_references_by_type(self, identifier: str) -> dict[str, list[str]]:
        """Return all the word-level references for an entity, grouped by type."""
        bcvwpreftypes = [
            "explicit_instances",
            "pronominal_referents",
            "subject_referents",
        ]
        return {
            reftype: references
            for reftype, references in self.references_by_type(identifier).items()
            if reftype in bcvwpreftypes
        }


class DeitiesReader(EntityReader):
    """Read Deities data."""

    acaidir = "deities"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiDeity


class GroupsReader(EntityReader):
    """Read Groups data."""

    acaidir = "groups"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiGroup


class PeopleReader(EntityReader):
    """Read People data."""

    acaidir = "people"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiPerson

    def preflabel(self, personid: str, language: str = "eng") -> str:
        if language != "eng":
            raise NotImplementedError("Not implemented for languages other than English.")
        preflabel: str = self.data[personid].preflabel(language)
        return preflabel

    def children_str(self, personid: str, language: str = "eng") -> str:
        """Return a string identifying children."""
        preflabel = self.preflabel(personid, language)
        personinst: AcaiPerson = self.data[personid]
        if personinst.offspring:
            childstr: str = ", ".join(
                [self.data[child].preflabel(language) for child in personinst.offspring]
            )
            if len(childstr) > 1:
                childrenstr = f"children of {preflabel} were"
            else:
                childrenstr = f"child of {preflabel} was"
            return f"The {childrenstr} {childstr}."
        else:
            return ""

    def partner_str(self, personid: str, language: str = "eng") -> str:
        """Return a string identifying partners."""
        preflabel = self.preflabel(personid, language)
        personinst: AcaiPerson = self.data[personid]
        if personinst.partners:
            spousestr: str = ", ".join(
                [self.data[partner].preflabel(language) for partner in personinst.partners]
            )
            if len(spousestr) > 1:
                partnerstr = f"partners of {preflabel} were"
            else:
                partnerstr = f"partner of {preflabel} was"
            return f"The {partnerstr} {spousestr}."
        else:
            return ""


class PlacesReader(EntityReader):
    """Read Places data."""

    acaidir = "places"
    jsondir = ACAIROOT / acaidir / "json"
    entryclass = AcaiPlace


# class FaunaReader(EntityReader):
#     """Read Places data."""

#     acaidir = "fauna"
#     jsondir = ACAIROOT / acaidir / "json"
#     entryclass = AcaiFaunaEntry



"""Utilities for working with MARBLE data."""

import re
from typing import Any
from warnings import warn


from biblelib.word import BCVID, BCVIDRange, BCVWPID, fromubs, simplify
from biblelib.unit import unitrange


def is_prot_marbleid(marbleid: str) -> bool:
    """Return True if a Marble ID from the Protestant canon."""
    if re.match("[A|G|H|L]", marbleid):
        # strip a leading language code if present
        marbleid = marbleid[1:]
    # check book indices
    bookindex = marbleid[:3]
    return (bookindex <= "039") or ("066" >= bookindex >= "040")


def from_marbleid(marbleid: str, omitdc: bool = True, verbose: bool = False) -> list[str]:
    """Return a list of BCVWP ID from a Marble ID like H00501900500034.

    With omitdc = True (the default), silently drop references outside
    the Protestant canon. With verbose = True (default is False), warn
    about these.

    """
    if re.match("[A|G|H|L]", marbleid):
        # strip a leading language code if present
        marbleid = marbleid[1:]
    # TODO: rehydrate a range into a sequence
    try:
        if omitdc and not is_prot_marbleid(marbleid):
            # drop references outside the Protestant canon
            if verbose:
                warn(f"Dropping DC reference: {marbleid}")
            return []
        else:
            return [bcv.ID for bcv in fromubs(marbleid)]
    except Exception as e:
        print(f"Failed on {marbleid}: {e}")
        return []


def from_marbleid_range(ref: str, verbose: bool = False) -> list[Any]:
    """Return a list of BCV(WP) IDs from a Marble ID that might be a range.

    If a range, it may produce a list of BCVID instances: Hebrew
    references may return multiple BCVWPID instances when MARBLE
    tokenization a word together that Macula splits. So items in the
    results list can be heterogeneous as to reference level.

    With verbose = True (default is False), warn about Deuterocanon
    references: otherwise drop them silently.

    """
    bcvlist: list[Any] = []
    # bcvlist: list[BCVID | BCVWPID] = []
    if "-" in ref:
        # range reference: rehydrate into a sequence
        # probably fragile assumptions here
        startmid, endmid = ref.split("-")
        start_bcvwpids = fromubs(startmid)
        end_bcvwpids = fromubs(endmid)
        try:
            if start_bcvwpids[0].chapter_ID == end_bcvwpids[0].chapter_ID:
                # this silently assumes the range is contiguous
                refrange = BCVIDRange(start_bcvwpids[0], end_bcvwpids[-1])
                bcvlist = refrange.enumerate()
            else:
                vrange = unitrange.VerseRange(start_bcvwpids[0], end_bcvwpids[-1])
                bcvlist = vrange.enumerate_ids()
            return bcvlist
        except Exception as e:
            print(f"Cannot process range {ref}, dropping\n{e}")
            return []
        # return unitrange.VerseRange(start=start_bcvwpid, end=end_bcvwpid)
    else:
        # silently drop DC references
        if is_prot_marbleid(ref):
            bcvwpref: list = fromubs(ref)
            # don't simplify. empty string if in DC
            return bcvwpref if bcvwpref else []
        else:
            if verbose:
                warn(f"Dropping {ref}")
            return []


def get_from_marbleids(seq: list, verbose: bool = False) -> list[str]:
    """Return a list of normalized BCV references from MARBLE ids.

    Map ranges to their enumerated items. Throughout convert to
    Biblelib-style BCVWPID token ids.

    """
    return [
        # no prefix: with part_index because Hebrew
        item.get_id()
        for marbleid in seq
        for item in from_marbleid_range(marbleid, verbose)
        if item
    ]



"""Base class for ACAI entities (people, places, etc.)"""

from dataclasses import asdict, dataclass, field
from enum import Enum

# TODO: enumerate a set of ACAI types


@dataclass
class Localization:
    """Localization data for an ACAI entity."""

    # values should be in acai.TargetLanguageCodes
    langcode: str
    preferred_label: str
    description: str
    alternate_labels: list = field(default_factory=list)
    # optional: where this information came from
    source: str = ""
    # optional: English gloss
    eng_gloss: str = ""


@dataclass
class LexicalRelation:
    """Manages a link to lexical content like UBSDGH/GNT."""

    # what kind of relation: Greek, Hebrew, other?
    # this should probably be an enumeration
    type: str
    # identifier for the lexical content
    lexicaluri: str


@dataclass
class EquivalenceRelation:
    """Manages a link to equivalent content describing this entity."""

    # what kind of relation: article, definition,
    # this should probably be an enumeration
    type: str
    # identifier for the lexical content
    # the namespace identifies the article source, like tynbd
    equivalenceuri: str


# To be deprecated
class ImageRelationType(str, Enum):
    """Enumerate types for image relations.

    Don't confuse this with the _subject_ of an image. This captures
    the image medium or type. Unlike e.g. Dublin Core `Image`, this
    does not include videos. The vast majority of UBS images are
    'photograph' or 'drawing', with an occaional 'painting'.

    Incorporates ideas from Schema.org's CreativeWork type, though not
    identical.

    """

    # "A picture or diagram made with a pencil, pen, or crayon rather than paint."
    DRAWING = "drawing"
    INFOGRAPHIC = "infographic"
    MAP = "map"
    PAINTING = "painting"
    PHOTOGRAPH = "photograph"
    # a la Logos
    FAMILYDIAGRAM = "family tree diagram"


class ImageType(str, Enum):
    """Enumerate types for image relations.

    Don't confuse this with the _subject_ of an image. This captures
    the image medium or type. Unlike e.g. Dublin Core `Image`, this
    does not include videos. The vast majority of UBS images are
    'photograph' or 'drawing', with an occaional 'painting'.

    Incorporates ideas from Schema.org's CreativeWork type, though not
    identical.

    """

    # "A picture or diagram made with a pencil, pen, or crayon rather than paint."
    DRAWING = "drawing"
    INFOGRAPHIC = "infographic"
    MAP = "map"
    PAINTING = "painting"
    PHOTOGRAPH = "photograph"
    # a la Logos
    FAMILYDIAGRAM = "family tree diagram"
    # also
    # TIMELINE
    # TABLE
    # SCULPTURE
    # WOODCUT
    # CARTOON: a drawing in a particular style
    # maybe later
    # 3D model
    # POSTER


# To be deprecated
@dataclass
class ImageRelation:
    """Manages a link to an image depicting this entity."""

    # what kind of image: photo, map, infographic, art, line drawing, family tree diagram
    # site plan, timeline
    # this should probably be an enumeration
    type: ImageRelationType
    # identifier for the lexical content
    # the namespace identifies the article source, like tynbd
    imageuri: str

    def asdict(self) -> dict:
        return {
            "type": self.type.value,
            "imageuri": self.imageuri,
        }


@dataclass
class BaseRecord:
    """Base class with common attributes for ACAI data records, including concepts.

    Other code should subclass this. This should not include type-specific attributes.
    """

    # <namespace>:<identifier>
    id: str
    # defined
    type: str = ""
    subtype: str = ""
    localizations: list[Localization] = field(default_factory=list)
    equivalence_relations: list[LexicalRelation] = field(default_factory=list)
    # list of BCVID/BCVWPID strings: select a few references that are
    # most important for the concept
    key_references: list = field(default_factory=list)
    # comprehensive list of references
    references: list = field(default_factory=list)


@dataclass
class Concept(BaseRecord):
    """Base class with common attributes for ACAI entities.

    Other code should subclass this. This doesn't include type-specific attributes.
    """

    # pointer to primary id. If this *is* the primary, it matches the
    # id field. This is for collecting several records together under
    # a main identifier.
    primary_id: str = ""
    # defined
    type: str = "person"  # for places, the only type is "place"
    lexical_relations: list[LexicalRelation] = field(default_factory=list)
    image_relations: list[ImageRelation] = field(default_factory=list)

    # # links to other entites or concepts that are fundamentally the same
    # same_as: list = field(default_factory=list)
    # possibly_same_as: list = field(default_factory=list)
    # # not sure about this
    # # referred_to_as: list = field(default_factory=list) # a list of acai_id pointing to respective entries
    # only_mentioned_in_apocrypha: bool = False  # "apocrypha" == NRSV apocrypha only
    # non_biblical: bool = False
    # lemmas: dict[str, list] = dataclasses.field(
    #     default_factory=dict
    # )  # lang: list<str> (list of lemmas in lang)
    # # will there be person types? roles?
    # # place_types: dict[str, list] = dataclasses.field(default_factory=dict) # key is source, list of types in sou

    def asdict(self) -> dict:
        """Return a dict for serialization."""
        return {
            "id": self.id,
            "primary_id": self.primary_id,
            "type": self.type,
            "localizations": [asdict(obj) for obj in self.localizations],
            "lexical_relations": [asdict(obj) for obj in self.lexical_relations],
            "equivalence_relations": [asdict(obj) for obj in self.equivalence_relations],
            "image_relations": [obj.asdict() for obj in self.image_relations],
            "key_references": self.key_references,
            "references": self.references,
        }



"""Parse UBSGNT into dataclasses.

The combination of an entryCode and a lemma makes for a unique
sense/meaning record (which also has an identifier, but it's opaque).

>>> from acai.lexicon import UBSParser
>>> ubsprs = UBSParser()
# these are Lexicon_Entry instances
>>> len(ubsprs)
5507
# these are LEXSense instances: 1-1 with LEXMeaning instances
>>> len(list(ubsprs.sense_iterator()))
9166


# lookup from a lemma
>>> pagis = ubsprs.get_lemma_entries("παγίς")
>>> pagis
<Lexicon_Entry: 003707000000000>
>>> pagis.get_meanings()
[<LEXMeaning: 003707001001000>, <LEXMeaning: 003707001002000>, <LEXMeaning: 003707001003000>]

# lookup from a Louw-Nida code
>>> trapsnare = ubsprs.get_entryCode_meanings("6.23")
>>> trapsnare
<LEXMeaning: 003707001001000>
>>> trapsnare.asdict()
{'identifier': '003707001001000', 'isBiblicalTerm': 'M', 'entryCode': '6.23', 'lexDomains': ['Artifacts'], 'lexSubDomains': ['Traps, Snares']}
# Dictionary for serialization
>>> trapsnare.serialize()
{'identifier': '003707001001000', 'isBiblicalTerm': 'M', 'lexDomains': 'Artifacts', 'lexSubDomains': 'Traps, Snares', 'entryCode': '6.23', 'lemma': 'παγίς', 'glosses': 'trap;snare', 'definitionShort': 'a device used for trapping or snaring, principally birds', 'lexReferences': 'LUK 21:35!2;ROM 11:9!9'}

# lookup from a strongs number: only one lemma for this one
>>> strongs_lemmas["G3803"]
[<Lexicon_Entry: 003707000000000>]

See LexiconEntry.py for additional details on the dataclasses.

# Aggregate data elements
>>> from acai.lexicon import Aggregator
>>> agg = Aggregator(ubsprs)
# mapping of domain names to subdomain names
>>> agg.domain_subdomains['Military Activities']
{'Soldiers, Officers', 'Prisoners of War', 'To Fight', 'Army', 'To Arm'}
# text attribute for lexSenses in a subdomain: default attribute is definitionShort
>>> agg.subdomain_words('To Fight')
['to go to war against', '(figurative extension of meaning of {L:μάχαιρα<SDBG:μάχαιρα:000000>}[a] ‘sword,’ {D:6.33})', 'to engage in open warfare', 'act of engaging in open warfare', '(figurative extension of meaning of {L:ῥομφαία<SDBG:ῥομφαία:000000>}[a] ‘broad sword,’ {D:6.32})', 'act of engaging in war or battle as a soldier', 'to engage in war or battle as a soldier', 'to oppose in battle']
# or the entryCode attribute (on the parten LexMeaning)
>>> agg.subdomain_words('To Fight', strfun=lambda lm: lm.lexSenses[0].parent.entryCode)
['55.2', '55.6', '55.5', '55.5', '55.6', '55.4', '55.4', '55.3']
# or the glosses
>>> agg.subdomain_words('To Fight', strfun=lambda lm: lm.lexSenses[0].glosses)
[['to rise up in arms against', 'to make war against'], ['war', 'fighting', 'conflict'], ['to wage war', 'war', 'fighting'], ['war', 'fighting'], ['war', 'fighting', 'conflict'], ['war', 'warfare'], ['to battle', 'to fight', 'to engage in war', 'warfare'], ['to meet in battle', 'to face in battle']]

"""

from collections import UserDict, defaultdict
from csv import DictWriter
from pathlib import Path
from typing import Any, Callable, Generator
import unicodedata

import requests
from lxml import etree

from biblelib.word import fromubs

from acai import DATAPATH

from .LexiconEntry import LEXSense, LEXMeaning, BaseForm, Lexicon_Entry


def get_ubsdgnt(langcode: str = "eng") -> str:
    """Return a string for the readable content from Github."""
    pathprefix: str = (
        "https://github.com/ubsicap/ubs-open-license/raw/main/dictionaries/greek/XML/UBSGreekNTDic-v1.0"
    )
    langcodes = {
        "eng": "en",
        "fra": "fr",
        "spa": "sp",
        # not ISO 639-3
        "zhs": "zhs",
        "zht": "zht",
    }
    return f"{pathprefix}-{langcodes[langcode]}.XML"


class UBSParser(UserDict):
    """Manage data read from UBSGNT."""

    # for lookup methods
    lemma_entries: dict[str, Lexicon_Entry] = {}
    entryCode_meanings: dict[str, LEXMeaning] = {}
    # sometimes multiple meanings for a word-level reference
    reference_meanings: dict[str, list[LEXMeaning]] = defaultdict(list)
    # for filtering: these domain codes are grammatical markers whose
    # semantics we may not care about
    grammatical_domains = {"90", "91", "92", "93"}

    def __init__(self, langcode: str = "eng"):
        """Initialize the parser."""
        super().__init__()
        # these store _elements_ for debugging
        # unfollowed import problem, so values are Any
        self.lexicon_entries: dict[str, Any] = {}
        self.baseform_elements: dict[str, Any] = {}
        self.lexmeaning_elements: dict[str, Any] = {}
        self.dicturi = get_ubsdgnt(langcode)
        r = requests.get(self.dicturi)
        assert r.status_code == 200, f"Failed to get content from {self.dicturi}"
        # normalize everything so matching works
        self.root = etree.fromstring(r.text)
        # iterate over Lexicon_Entry elements
        # for Greek, each Lexicon_Entry.lemma is unique, so could use
        # lemma for keys rather than identifier. Not sure about Hebrew though
        self.data = {
            thislexe.identifier: thislexe
            for lexentry in self.root.xpath("//Lexicon_Entry")
            if (thislexe := self._do_lexicon_entry(lexentry))
        }

    @staticmethod
    def _first_if(xpathexpr: str) -> str:
        """Return first string value in the xpath result, or empty string."""
        return xpathexpr[0] if len(xpathexpr) > 0 else ""

    # type complaint with this
    #    def _do_senses(self, lexmean: etree._Element) -> list[LEXSense]:
    def _do_senses(self, lexmean: Any) -> list[LEXSense]:
        """Process LEXSense content.

        Returns a list with a single LEXSense instance. This leaves
        room for other data that might have multiple.

        """
        # only ever one LEXSense for Greek
        lexsense = lexmean.xpath("LEXSenses/LEXSense")[0]
        lsdict = dict(
            languageCode=lexsense.xpath("@LanguageCode")[0],
            glosses=lexsense.xpath("Glosses/Gloss/text()"),
            definitionShort=self._first_if(lexsense.xpath("DefinitionShort/text()")),
            definitionLong=self._first_if(lexsense.xpath("DefinitionLong/text()")),
            comments=lexsense.xpath("Comments/text()"),
        )
        return [LEXSense(**lsdict)]

    # def _do_meaning(self, lexmean: etree._Element) -> LEXMeaning:
    def _do_meaning(self, lexmean: Any) -> LEXMeaning:
        """Process LEXMeaning contents."""
        # stash for debugging
        lexmean_id: str = lexmean.xpath("@Id")[0]
        self.lexmeaning_elements[lexmean_id] = lexmean
        try:
            lexsenses = self._do_senses(lexmean)
        except Exception as e:
            print(f"Failed on {lexmean}, {lexmean_id}\n{e}")
            lexsenses = []
        try:
            lexReferences = [
                ubsref.get_id()
                for ref in lexmean.xpath("LEXReferences/LEXReference/text()")
                for ubsref in fromubs(ref)
            ]
        except Exception as e:
            print(f"Failed on references for {lexmean}, {lexmean_id}\n{e}")
            lexReferences = []
        lexmeaning = LEXMeaning(
            identifier=lexmean_id,
            isBiblicalTerm=self._first_if(lexmean.xpath("@IsBiblicalTerm")),
            entryCode=self._first_if(lexmean.xpath("@EntryCode")),
            # does this capture multiple values correctly?
            lexDomains=lexmean.xpath("LEXDomains/LEXDomain/text()"),
            lexSubDomains=lexmean.xpath("LEXSubDomains/LEXSubDomain/text()"),
            lexSenses=lexsenses,
            lexReferences=lexReferences,
        )
        for lexs in lexsenses:
            lexs.parent = lexmeaning
        return lexmeaning

    def _do_baseform(self, baseformel: Any) -> BaseForm:
        """Process BaseForm contents."""
        # stash for debugging
        baseform_id: str = baseformel.xpath("@Id")[0]
        self.baseform_elements[baseform_id] = baseformel
        baseform = BaseForm(
            identifier=baseform_id,
            partsofspeech=baseformel.xpath("PartsOfSpeech/PartOfSpeech/text()"),
            lexMeanings=[
                self._do_meaning(lm)
                for lm in baseformel.xpath("LEXMeanings/LEXMeaning")
                if self._first_if(lm.xpath("@IsBiblicalTerm")) != "N"
            ],
        )
        for lm in baseform.lexMeanings:
            lm.parent = baseform
        return baseform

    def _do_lexicon_entry(self, lexentry: Any) -> Lexicon_Entry:
        """Process Lexicon_Entry contents."""
        lexentry_id: str = lexentry.xpath("@Id")[0]
        self.lexicon_entries[lexentry_id] = lexentry
        # PROCESS _do_baseform here
        thislexe = Lexicon_Entry(
            identifier=lexentry_id,
            lemma=unicodedata.normalize("NFKC", lexentry.xpath("@Lemma")[0]),
            strongCodes=lexentry.xpath("StrongCodes/Strong/text()"),
            # this assumes a single BaseForm per lexentry
            baseforms=[self._do_baseform(bf) for bf in lexentry.xpath("BaseForms/BaseForm")],
        )
        for bf in thislexe.baseforms:
            bf.parent = thislexe
        return thislexe

    # accessors
    # only build these on demand

    # just make lemma the key for self.data?
    def get_lemma_entries(self, lemma: str) -> Lexicon_Entry:
        """Return the Lexicon_Entry instance for lemma."""
        if not self.lemma_entries:
            self.lemma_entries = {entry.lemma: entry for entry in self.data.values()}
        return self.lemma_entries[lemma]

    def get_entryCode_meanings(self, entryCode: str) -> LEXMeaning:
        """Return the LEXMeaning instance for an entryCode."""
        if not self.entryCode_meanings:
            self.entryCode_meanings = {
                meaning.entryCode: meaning
                for entry in self.data.values()
                for baseform in entry.baseforms
                for meaning in baseform.lexMeanings
            }
        return self.entryCode_meanings[entryCode]

    def get_reference_meanings(self, reference: str) -> list[LEXMeaning]:
        """Return the LEXMeaning instances for an reference."""
        if not self.reference_meanings:
            for entry in self.data.values():
                for baseform in entry.baseforms:
                    for meaning in baseform.lexMeanings:
                        for ref in meaning.lexReferences:
                            self.reference_meanings[ref].append(meaning)
        return self.reference_meanings[reference]

    def get_strongs_lemmas(self) -> dict[str, list[Lexicon_Entry]]:
        """Return a mapping from strongs to lemmas."""
        strongs_lemmas: dict[str, list[Lexicon_Entry]] = defaultdict(list)
        for entry in self.data.values():
            for strong in entry.strongCodes:
                strongs_lemmas[strong].append(entry)
        return strongs_lemmas

    def get_strongs_meanings(self) -> dict[str, list[LEXMeaning]]:
        """Return a mapping from strongs to lexMeanings."""
        strongs_lemmas: dict[str, list[LEXMeaning]] = defaultdict(list)
        for entry in self.data.values():
            for baseform in entry.baseforms:
                for meaning in baseform.lexMeanings:
                    for strong in entry.strongCodes:
                        strongs_lemmas[strong].append(meaning)
        return strongs_lemmas

    def meaning_iterator(self) -> Generator[LEXMeaning, None, None]:
        """Return a generator over senses."""
        for lexent in self.values():
            for baseform in lexent.baseforms:
                for lexmean in baseform.lexMeanings:
                    yield lexmean

    def sense_iterator(self) -> Generator[LEXSense, None, None]:
        """Return a generator over senses.

        Note these are 1-1 with meanings for DGNT.
        """
        for lexmean in self.meaning_iterator():
            for lexsense in lexmean.lexSenses:
                yield lexsense

    def sense_references(self) -> dict[str, list[str]]:
        """Return a dict of sense identifiers and their references."""
        return {
            parent.identifier: parent.lexReferences
            for sense in self.sense_iterator()
            if (parent := sense.parent)
        }

    def write_sense_data(self, outpath: Path = DATAPATH / "lexicon/sensedata.tsv") -> None:
        """Write out sense data as TSV."""
        self.definitionShort = {}
        self.comments = {}
        self.glosses = {}
        fieldnames = [
            "LEXMeaningID",
            "LEXMeaningEntryCode",
            "glosses",
            "definitionShort",
            "comments",
        ]
        with outpath.open("w", encoding="utf-8", newline="") as f:
            writer = DictWriter(
                f,
                fieldnames=fieldnames,
                delimiter="\t",
            )
            writer.writeheader()
            for lexsense in self.sense_iterator():
                sensedict = lexsense.asdict()
                del sensedict["languageCode"]
                del sensedict["definitionLong"]
                parent = lexsense.parent
                lexmeaningid = parent.identifier
                sensedict.update({"LEXMeaningID": lexmeaningid})
                sensedict.update({"LEXMeaningEntryCode": parent.entryCode})
                writer.writerow(sensedict)
                self.definitionShort[lexmeaningid] = lexsense.definitionShort
                self.comments[lexmeaningid] = lexsense.comments
                self.glosses[lexmeaningid] = lexsense.glosses


# ordering here may be suboptimal
class Aggregator:
    """Aggregate various data elements."""

    def __init__(self, parser: UBSParser) -> None:
        """Initialize the aggregator."""
        self.parser = parser
        # maps domain strings to their subdomain strings
        # might be nicer to also capture the number?
        self.domain_subdomains: dict[str, set[str]] = defaultdict(set)
        # maps subdomain strings to their LEXMeaning instances
        self.subdomain_meanings: dict[str, list[LEXMeaning]] = defaultdict(list)
        for meaning in self.parser.meaning_iterator():
            # there's really only one for DGNT
            for subdomain in meaning.lexSubDomains:
                self.subdomain_meanings[subdomain].append(meaning)
            for domain in meaning.lexDomains:
                self.domain_subdomains[domain].add(subdomain)

    def subdomain_words(
        self, subdomain: str, strfun: Callable = lambda lm: lm.lexSenses[0].definitionShort
    ) -> list[list[str]]:
        """Return a list of lists of words for a subdomain.

        strfun is a function that takes a LEXMeaning and returns a
        string. Default is to return definitionShort for the first
        (only) lexSense.

        """
        return [strfun(meaning) for meaning in self.subdomain_meanings[subdomain]]



"""Utilities for working with Strong's numbers.

This duplicates functionality in
internal-alignments/src/util/strong.py. Should eventually refactor
into a single library to avoid this, but until then, keep in sync

Obligatory disclaimer: Strong's numbers are a mess, and we shouldn't
rely on them. Nevertheless, some data comes with them, so we should
maintain them, and (when forced) use them.

"""

import re


def normalize_strongs(strongs: str | int, prefix: str = "") -> str:
    """Return a normalized Strongs id."""
    _strongsre: re.Pattern = re.compile(r"[AGH]?\d{1,4}[a-d]?")
    specials: dict[str, str] = {
        "1537+4053": "G4053b",
        "5228+1537+4053": "G4053c",
        "1417+3461": "G3461b",
    }
    # special case for uW KeyTerms data: some like G29620. It appears
    # the last digit is always zero
    if isinstance(strongs, str) and strongs.startswith("G") and len(strongs) == 6:
        if strongs.endswith("0"):
            strongs = strongs[:-1]
        else:
            raise ValueError(f"6-char Strong's code: {strongs}")
    # some special cases for SBLGNT data
    if strongs in specials:
        normed = specials[str(strongs)]
    elif isinstance(strongs, int):
        assert prefix, f"prefix is required for a bare int: {strongs}"
        normed = f"{prefix}{strongs:0>4}"
    elif _strongsre.fullmatch(strongs):
        # check for initial prefix: save if available
        if re.match(r"[AGH]", strongs):
            if prefix:
                print(f"Overwriting prefix parameter {prefix} for {strongs}")
            prefix = strongs[:1]
        base = re.sub(r"\D", "", strongs)
        # final letter
        if re.search("[a-d]$", strongs):
            suffix = strongs[-1]
        else:
            suffix = ""
        # might need other tests here
        # may leave prefix unspecified
        assert prefix, f"prefix must be specified: {strongs}"
        normed = f"{prefix}{base:0>4}{suffix}"
    else:
        raise ValueError(f"Invalid Strong's code: {strongs}")
    return normed
